'''
Module with Reader class used to read weights to normalize between inclusive samples
'''

import os
from functools import lru_cache

import pandas    as pnd
from dmu.logging.log_store import LogStore
from fitter                import pdg_utils as pu

log = LogStore.add_logger('rx_fitter:inclusive_sample_weights')
#---------------------------
class Reader:
    '''
    Class used to add weights that normalize inclusive samples
    '''
    #---------------------------
    def __init__(self, df : pnd.DataFrame):
        self._df      = df
        self._fu      = 0.408
        self._fs      = 0.100

        self._bu_proc = 'Bu_JpsiX_ee_eq_JpsiInAcc'
        self._bd_proc = 'Bd_JpsiX_ee_eq_JpsiInAcc'
        self._bs_proc = 'Bs_JpsiX_ee_eq_JpsiInAcc'
    #---------------------------
    @lru_cache(maxsize=10)
    def _get_br_wgt(self, proc : str) -> float:
        '''
        Will return ratio:

        decay file br / pdg_br
        '''

        #--------------------------------------------
        #Decay B+sig
        #0.1596  MyJ/psi    K+           SVS ;
        #--------------------------------------------
        #Decay B0sig
        #0.1920  MyJ/psi    MyK*0        SVV_HELAMP PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus ;
        #--------------------------------------------
        #Decay B_s0sig
        #0.1077  MyJ/psi    Myphi        PVV_CPLH 0.02 1 Hp pHp Hz pHz Hm pHm;
        #--------------------------------------------

        if proc == self._bu_proc:
            return pu.get_bf('B+ --> J/psi(1S) K+') / 0.1596

        if proc == self._bd_proc:
            return pu.get_bf('B0 --> J/psi(1S) K0') / 0.1920

        if proc == self._bs_proc:
            return pu.get_bf('B_s()0 --> J/psi(1S) phi') / 0.1077

        raise ValueError(f'Invalid process {proc}')
    #---------------------------
    @lru_cache(maxsize=10)
    def _get_hd_wgt(self, proc : str) -> float:
        '''
        Will return hadronization fractions used as weights
        '''
        log.info(f'Getting hadronization weights for sample {proc}')

        if proc in [self._bu_proc, self._bd_proc]:
            return self._fu

        if proc == self._bs_proc:
            return self._fs

        raise ValueError(f'Invalid process: {proc}')
    #---------------------------
    def _get_stats(self, path):
        proc = os.path.basename(path).replace('.json', '')
        df   = pnd.read_json(path)

        return proc, df
    #---------------------------
    def _good_rows(self, r1 : pnd.Series, r2 : pnd.Series) -> bool:
        if {r1.Polarity, r2.Polarity} != {'MagUp', 'MagDown'}:
            log.error('Latest rows are not of opposite polarities')
            return False

        if r1.Events <= 0 or r2.Events <= 0:
            log.error('Either polarity number of events is negative')
            return False

        return True
    #---------------------------
    @lru_cache(maxsize=10)
    def _get_st_wgt(self, proc : str) -> float:
        return 1
    #---------------------------
    def _get_weight(self, row : pnd.Series) -> float:
        w1 = self._get_st_wgt(row.proc)
        w2 = self._get_hd_wgt(row.proc)
        w3 = self._get_br_wgt(row.proc)

        return w1 * w2 * w3
    #---------------------------
    def get_weights(self) -> pnd.Series:
        '''
        Returns:

        Pandas series with sample weights
        '''
        sr_wgt = self._df.apply(self._get_weight, axis=1)

        return sr_wgt
#---------------------------



'''
Module holding PChain class
'''
# pylint: disable=invalid-name
# pylint: disable=missing-function-docstring

from dmu.logging.log_store import LogStore

log=LogStore.add_logger('fitter:pchain')
#----------------------------------
class PChain:
    '''
    Class meant to represent a decay chain
    '''
    #----------------------------------
    def __init__(self, pid, mid, gmid, ggmid):
        self._TID          = pid
        self._MOTHER_TID   = mid
        self._GMOTHER_TID  = gmid
        self._GGMOTHER_TID = ggmid
    #----------------------------------
    def ID(self):
        return self._TID

    def MID(self):
        return self._MOTHER_TID

    def GMID(self):
        return self._GMOTHER_TID

    def GGMID(self):
        return self._GGMOTHER_TID
    #----------------------------------------------------
    def MatchDecay(self, l_dec_id):
        if len(l_dec_id) == 1:
            return self._TID ==      l_dec_id[0]
        if len(l_dec_id) == 2:
            return  self._TID == abs(l_dec_id[0]) and self._MOTHER_TID == abs(l_dec_id[1])
        if len(l_dec_id) == 3:
            return  self._TID == abs(l_dec_id[0]) and self._MOTHER_TID == abs(l_dec_id[1]) and self._GMOTHER_TID == abs(l_dec_id[2])
        if len(l_dec_id) == 4:
            return  self._TID == abs(l_dec_id[0]) and self._MOTHER_TID == abs(l_dec_id[1]) and self._GMOTHER_TID == abs(l_dec_id[2]) and self._GGMOTHER_TID == abs(l_dec_id[3])

        return False
    #----------------------------------------------------
    def HasInChain(self, ID):
        _return = False

        if self._MOTHER_TID   == ID:
            _return = True
        if self._GMOTHER_TID  == ID:
            _return = True
        if self._GGMOTHER_TID == ID:
            _return = True

        return _return
    #----------------------------------------------------
    def MatchUpstream(self, IDFirstDau, HeadPart):
        _return = False
        if  self._MOTHER_TID   == IDFirstDau and self._GMOTHER_TID  == HeadPart:
            _return = True
        if  self._GMOTHER_TID  == IDFirstDau and self._GGMOTHER_TID == HeadPart:
            _return = True
        if  self._GGMOTHER_TID == IDFirstDau:
            _return = True

        return _return
    #----------------------------------------------------
    def MatchID(self, iD):
        return self._TID == abs(iD)

    def MatchMother(self, iD):
        return self._MOTHER_TID == abs(iD)

    def MatchGMother(self, iD):
        return self._GMOTHER_TID == abs(iD)

    def MatchGGMother(self, iD):
        return self._GGMOTHER_TID == abs(iD)
#----------------------------------------------------



'''
Module containing class ModelScales and helper functions
'''

import math
import numpy
import pandas   as pnd
import jacobi   as jac

from dmu.logging.log_store                 import LogStore
from dmu.generic                           import hashing
from dmu.generic                           import utilities  as gut
from rx_selection                          import selection  as sel
from rx_efficiencies.decay_names           import DecayNames as dn
from rx_efficiencies.efficiency_calculator import EfficiencyCalculator

log=LogStore.add_logger('fitter:prec_scales')
#------------------------------------------
class PrecScales:
    '''
    Class used to calculate scale factor between yields of partially reconstructed component and signal
    '''
    #------------------------------------------
    def __init__(self, proc : str, q2bin : str):
        '''
        proc : Nickname of decay process, nicknames are in the DecayNames class
        q2bin: Needed to apply correct selection to get correct efficiencies and scales
        '''
        self._proc        = proc
        self._q2bin       = q2bin

        self._d_frbf      : dict
        self._trigger     = 'Hlt2RD_BuToKpEE_MVA'
        self._initialized = False

        self._hash        = self._get_hash()
    #------------------------------------------
    def _get_hash(self) -> str:
        process = dn.sample_from_decay(self._proc)
        d_sel   = sel.selection(trigger=self._trigger, q2bin=self._q2bin, process=process)
        hsh     = hashing.hash_object([self._proc, self._q2bin, d_sel])

        return hsh
    #------------------------------------------
    def _check_arg(self, l_val, val, name):
        if val not in l_val:
            raise ValueError(f'{name} {val} not allowed')

        log.debug(f'{name:<20}{"->":20}{val:<20}')
    #------------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        log.debug('Initializing')
        self._d_frbf = gut.load_data(package='rx_efficiencies_data', fpath='scales/fr_bf.yaml')

        self._initialized = True
    #------------------------------------------
    def _get_fr(self, proc : str) -> float:
        '''
        Returns hadronization fraction for given process
        '''
        if   proc.startswith('bp'):
            fx = 'fu'
        elif proc.startswith('bd'):
            fx = 'fd'
        elif proc.startswith('bs'):
            fx = 'fs'
        else:
            raise ValueError(f'Cannot find hadronization fraction for: {proc}')

        fx = self._d_frbf['fr'][fx]

        return fx
    #------------------------------------------
    def _mult_brs(self, l_br : list[tuple[float,float]]) -> tuple[float,float]:
        '''
        Parameters
        -----------------------
        l_br: List of branching fraction, branching fraction error tuples

        Returns
        -----------------------
        A tuple with the product of those branching fractions and the error in the product
        '''
        log.debug('Multiplying branching fractions')

        l_br_val = [ float(br[0]) for br in l_br ] # These numbers come from YAML files
        l_br_err = [ float(br[1]) for br in l_br ] # when using "e" in scientific notation, these numbers are made into strings

        br_cov   = numpy.diag(l_br_err) ** 2
        val, var = jac.propagate(
            math.prod,
            l_br_val,
            br_cov)

        err      = math.sqrt(var)
        val      = float(val)

        return val, err
    #------------------------------------------
    def _get_br(self, proc : str) -> tuple[float,float]:
        log.debug(f'Calculating BR for {proc}')

        l_dec = dn.subdecays_from_nickname(proc)
        l_bf  = [ self._d_frbf['bf'][dec] for dec in l_dec ]

        return self._mult_brs(l_bf)
    #------------------------------------------
    def _get_ef(self, proc : str) -> tuple[float,float]:
        '''
        Parameters
        --------------
        proc: Nickname to process, e.g. bpkpee

        Returns
        --------------
        Tuple with efficiency value and error
        '''
        sample = dn.sample_from_decay(proc)

        log.debug(f'Calculating efficiencies for {sample}')
        obj = EfficiencyCalculator(q2bin=self._q2bin)
        val = obj.get_efficiency(sample=sample)

        return val
    #------------------------------------------
    def _print_vars(self, l_tup : list[tuple[float,float]], proc : str) -> None:
        log.debug('')
        log.debug(f'Decay: {proc}')
        log.debug('-' * 20)
        log.debug(f'{"Var":<20}{"Value":<20}{"Error":<20}')
        log.debug('-' * 20)
        for (val, err), name in zip(l_tup, ['fr', 'br', 'eff']):
            log.debug(f'{name:<20}{val:<20.3e}{err:<20.3e}')
        log.debug('-' * 20)
    #------------------------------------------
    def get_scale(self, signal : str) -> tuple[float,float]:
        '''
        Parameters
        -----------------------
        signal: String representing signal WRT which to scale, e.g. bpkpee

        Returns
        ---------------
        Scale factor k and error, meant to be used in:

        Nprec = k * Nsignal

        reparametrization, during fit.
        '''
        self._initialize()

        fr_bkg = self._get_fr(self._proc)
        br_bkg = self._get_br(self._proc)
        ef_bkg = self._get_ef(self._proc)

        fr_sig = self._get_fr(signal)
        br_sig = self._get_br(signal)
        ef_sig = self._get_ef(signal)

        l_tup = [fr_sig, br_sig, ef_sig, fr_bkg, br_bkg, ef_bkg]
        l_val = [ tup[0] for tup in l_tup]
        l_err = [ tup[1] for tup in l_tup]
        cov   = numpy.diag(l_err) ** 2

        self._print_vars(l_tup[:3], proc=    signal)
        self._print_vars(l_tup[3:], proc=self._proc)

        val, var = jac.propagate(lambda x : (x[3] * x[4] * x[5]) / (x[0] * x[1] * x[2]), l_val, cov)
        val = float(val)
        err = math.sqrt(var)

        return val, err
#------------------------------------------



'''
This module contains BaseFitter
'''
from typing                   import cast
import matplotlib.pyplot as plt

from omegaconf                import OmegaConf, DictConfig
from dmu.stats.fitter         import Fitter
from dmu.stats.zfit_plotter   import ZFitPlotter
from dmu.generic              import utilities  as gut
from dmu.stats                import utilities  as sut
from dmu.logging.log_store    import LogStore
from rx_selection             import selection  as sel
from zfit.result              import FitResult  as zres
from zfit.core.interfaces     import ZfitData   as zdata
from zfit.core.interfaces     import ZfitPDF    as zpdf

log=LogStore.add_logger('fitter:base_fitter')
# ------------------------
class BaseFitter:
    '''
    Fitting base class, meant to

    - Provide basic functionality to fiters for data and simulation
    - Behave as a dependency sink, avoiding circular imports
    '''
    # ------------------------
    def __init__(self):
        '''
        Used to hold attributes passed from derived classes
        '''
        self._sample  : str = ''
        self._trigger : str = ''
        self._project : str = ''
        self._q2bin   : str = ''
    # ------------------------
    def _fit(
            self,
            cfg   : DictConfig,
            data  : zdata,
            model : zpdf,
            d_cns : dict[str,tuple[float,float]]|None = None) -> zres:
        '''
        Parameters
        --------------------
        cfg  : Fitting configuration
        data : Zfit data object
        model: Zfit PDF
        d_cns: Dictionary mapping parameter names to tuples of value and error
               This is needed to apply constraints to fit

        Returns
        --------------------
        DictConfig object with parameters names, values and errors
        '''
        fit_cfg = OmegaConf.to_container(cfg, resolve=True)
        fit_cfg = cast(dict, fit_cfg)

        if d_cns is not None:
            fit_cfg['constraints'] = d_cns

        ftr = Fitter(pdf=model, data=data)
        res = ftr.fit(cfg=fit_cfg)

        return res
    # ------------------------
    def _get_sensitivity(self, res : zres|None) -> float:
        '''
        Parameters
        --------------
        res: Result object from fit

        Returns
        --------------
        fit sensitivity in %
        '''
        if res is None:
            log.debug('Missing result object, cannot get sensitivity')
            return -1

        cres = sut.zres_to_cres(res=res)

        if 'nsignal' not in cres:
            log.debug('Missing nsig entry, cannot get sensitivity')
            return -1

        value = cres['nsignal']['value']
        error = cres['nsignal']['error']

        return 100 * error / value
    # --------------------------
    def _brem_cuts_from_cuts(self, cuts : dict[str,str]) -> str:
        '''
        Parameters
        --------------
        cuts: Dictionary with cuts used for fit

        Returns
        --------------
        String with brem requirements
        '''
        l_brem_cut = []
        for cut in cuts.values():
            if 'nbrem' not in cut:
                continue
            l_brem_cut.append(cut)

        brem_cuts = '; '.join(l_brem_cut)

        return brem_cuts
    # --------------------------
    def _get_selection_text(self, cuts : dict[str,str]) -> tuple[str,str]:
        '''
        Parameters
        --------------
        cuts: Dictionary with cuts used for fit

        Returns
        --------------
        Tuple with:

        - Multiple lines with cuts that were used for fit, but are not default, plus MVA cut
        - Brem categories choice
        '''
        # For components like combinatorial, there is no MC sample
        # Therefore the selection or brem category does not make sense
        if self._sample == 'NA':
            return '', ''

        brem_cuts = self._brem_cuts_from_cuts(cuts=cuts)
        # Pick default selection
        with sel.custom_selection(d_sel={}):
            d_sel_def = sel.selection(
                process=self._sample,
                trigger=self._trigger,
                q2bin  =self._q2bin)

        l_expr = []
        # Collect all the cuts that are different
        # from default selection
        for name, expr in cuts.items():
            if name not in d_sel_def:
                expr = cuts[name]
                l_expr.append(expr)
                continue

            def_expr = d_sel_def[name]
            if expr != def_expr:
                l_expr.append(expr)

        # Remove differences in brem, will be done separately
        l_expr_no_brem = [ expr for expr in l_expr if 'nbrem' not in expr ]
        new_cuts       = '\n'.join(l_expr_no_brem)

        return new_cuts, brem_cuts
    # --------------------------
    def _entries_from_data(self, data : zdata) -> int:
        '''
        Parameters
        ---------------
        data: Dataset used in the fit

        Returns
        ---------------
        Number of entries in data that were used for the fit,
        which are in the fit observable range
        '''
        obs          = data.space
        [minx, maxx] = sut.range_from_obs(obs=obs)

        arr_mass = data.to_numpy()
        mask     = (minx < arr_mass) & (arr_mass < maxx)
        arr_mass = arr_mass[mask]
        nentries = len(arr_mass)

        return nentries
    # --------------------------
    def _get_text(
            self,
            data : zdata,
            res  : zres|None,
            cuts : dict[str,str]) -> tuple[str,str]:
        '''
        Parameters
        --------------
        data: Zfit data used for fit
        res : zfit result object
        cuts: Dictionary with cuts used to get data

        Returns
        --------------
        Tuple with:

        - Title for fit plot
        - Text that goes inside plot with selection information
        '''
        nentries          = self._entries_from_data(data=data)
        sel_txt, brem_txt = self._get_selection_text(cuts=cuts)

        sensitivity = self._get_sensitivity(res=res)
        title       = f'$\\delta={sensitivity:.2f}$%; Entries={nentries:.0f}; Brem:{brem_txt}'

        return title, sel_txt
    # ------------------------
    def _save_fit(
            self,
            cuts     : dict[str,str],
            cfg      : DictConfig,
            out_path : str,
            model    : zpdf|None,
            res      : zres|None,
            data     : zdata,
            d_cns    : dict[str,tuple[float,float]]|None=None) -> None:
        '''
        Parameters
        --------------
        cuts     : Selection used for fit
        cfg      : Plotting configuration
        out_path : Directory where fit will be saved
        model    : PDF from fit, can be None if dataset was empty
        res      : Zfit result object, can be None if fit was to get a KDE
        data     : data from fit
        d_cns    : Dictionary mapping parameter name to value error tuple.
                   Used for constraining that parameter
        '''
        plt_cfg = OmegaConf.to_container(cfg, resolve=True)
        plt_cfg = cast(dict, plt_cfg)

        # If no entries were present
        # There will not be PDF
        title, text         = self._get_text(data=data, res=res, cuts=cuts)
        plt_cfg['title'   ] = title
        plt_cfg['ext_text'] = text

        if model is not None:
            ptr = ZFitPlotter(data=data, model=model)
            ptr.plot(**plt_cfg)
        else:
            plt.figure()

        sel_path = f'{out_path}/selection.yaml'
        gut.dump_json(cuts, sel_path)

        sut.save_fit(
            data   = data,
            model  = model,
            res    = res,
            d_const= d_cns,
            fit_dir= out_path)
# ------------------------



'''
Module containing PRec
'''
import os
import copy
import json
import pprint
from typing     import Union
from contextlib import contextmanager

import numpy
import slugify
import pandas            as pnd
import matplotlib.pyplot as plt

from omegaconf              import OmegaConf
from dmu.stats.zfit         import zfit
from dmu.generic            import hashing
from dmu.logging.log_store  import LogStore
from dmu.stats.zfit_plotter import ZFitPlotter
from dmu.stats.utilities    import is_pdf_usable
from dmu.stats              import utilities as sut
from dmu.workflow.cache     import Cache

from zfit.core.parameter   import Parameter as zpar
from zfit.core.basepdf     import BasePDF   as zpdf
from zfit.core.interfaces  import ZfitSpace as zobs
from rx_selection          import selection as sel
from rx_data.rdf_getter    import RDFGetter
from ROOT                  import RDataFrame

from fitter.inclusive_decays_weights import Reader as inclusive_decays_weights
from fitter.inclusive_sample_weights import Reader as inclusive_sample_weights

log=LogStore.add_logger('fitter:prec')
#-----------------------------------------------------------
class PRec(Cache):
    '''
    Class used to calculate the PDF associated to the partially reconstructed background
    '''
    use_cache = True # Use cached if found
    #-----------------------------------------------------------
    def __init__(
        self,
        samples  : list[str],
        trig     : str,
        q2bin    : str,
        d_weight : dict[str,int],
        out_dir  : str = ''):
        '''
        Parameters:
        -------------------------
        samples  : MC samples
        trig     : HLT2 trigger.
        q2bin    : q2 bin
        d_weight : Dictionary specifying which weights to use, e.g. {'dec' : 1, 'sam' : 1}
        out_dir  : Directory where cached outputs will go WRT _cache_root, default empty
        '''
        self._l_sample = samples
        self._trig     = trig
        self._q2bin    = q2bin
        self._d_wg     = copy.deepcopy(d_weight)

        self._name     : str
        self._d_fstat  = {}
        d_rdf, uid     = self._get_samples_rdf()
        self._d_rdf    = d_rdf

        self._d_match         = self._get_match_str()
        self._l_mass          = ['B_Mass', 'B_Mass_smr', 'B_const_mass_M', 'B_const_mass_psi2S_M']
        self._min_entries     = 40 # Will not build KDE if fewer entries than this are found
        self._min_isj_entries = 500 #if Fewer entries than this, switch from ISJ to FFT

        self._check_valid(self._q2bin, ['low', 'central', 'jpsi', 'psi2', 'high'], 'q2bin')
        self._check_weights()

        # This should be usable to make hashes
        # backlashes and dollar signs are not hashable in strings
        d_hash_match = { slugify.slugify(ltex) : value for ltex, value in self._d_match.items() }

        super().__init__(
            out_path = out_dir,
            uid      = uid,
            d_wg     = OmegaConf.to_container(d_weight, resolve=True),
            d_match  = d_hash_match)
    #-----------------------------------------------------------
    def _get_df(self) -> dict[str,pnd.DataFrame]:
        '''
        Returns
        -------------------
        Dictionary where:

        Key  : Identifier for ccbar component, e.g. "$B_d\to c\bar{c}(\to ee)H_s$"
        Value: Dictionary with:
            Key  : Name of component
            Value: Dataframe with entries for only that component
        '''
        d_df = self._get_samples_df()
        d_df = { sample : self._add_dec_weights(sample, df) for sample, df in d_df.items() }
        df   = pnd.concat(d_df.values(), axis=0)
        df   = self._add_sam_weights(df)

        if len(df) == 0:
            return {}

        arr_wgt      = df.wgt_dec.to_numpy() * df.wgt_sam.to_numpy()
        df['wgt_br'] = self._normalize_weights(arr_wgt)

        d_df = { component : df.query(cut) for component, cut in self._d_match.items() }

        return d_df
    #-----------------------------------------------------------
    def _need_var(self, name : str) -> bool:
        needed = False

        if name.endswith('ID'):
            needed = True

        if name in self._l_mass:
            needed = True

        return needed
    #-----------------------------------------------------------
    def _filter_rdf(
        self,
        rdf    : RDataFrame,
        uid    : str,
        sample : str) -> tuple[RDataFrame,str]:
        '''
        Parameters
        -----------------
        rdf    : ROOT dataframe before selection
        uid    : Unique identifier of dataframe
        sample : Sample for which selection is done, e.g. Bu_JpsiX...

        Returns
        -----------------
        Tuple with:

        - ROOT dataframe after selection
        - Updated Unique identifier that takes into account the selection
        '''
        d_sel         = sel.selection(trigger=self._trig, q2bin=self._q2bin, process=sample)
        d_sel['mass'] = '(1)'
        for name, expr in d_sel.items():
            rdf = rdf.Filter(expr, name)

        uid = hashing.hash_object([uid, d_sel])

        return rdf, uid
    #-----------------------------------------------------------
    def _get_samples_rdf(self) -> tuple[dict[str,RDataFrame],str]:
        '''
        IMPORTANT: This method has to run dataframe creation lazily

        Returns
        -----------------
        Tuple of 2 elements:

        - Dictionary with
            - Key: Name of the ccbar sample
            - Value: ROOT dataframe after the selection

        - Concatenation of unique identifiers
        '''
        d_rdf    = {}
        full_uid = ''
        for sample in self._l_sample:
            gtr        = RDFGetter(sample=sample, trigger=self._trig)
            rdf        = gtr.get_rdf()
            uid        = gtr.get_uid()
            rdf, uid   = self._filter_rdf(rdf=rdf, sample=sample, uid=uid)

            d_rdf[sample] = rdf
            full_uid     += uid

        return d_rdf, full_uid
    #-----------------------------------------------------------
    def _get_samples_df(self) -> dict[str,pnd.DataFrame]:
        '''
        Returns
        ------------------
        Dictionary with:

        - Key: Name of ccbar sample
        - Value: Pandas dataframe with only the needed columns
        '''
        d_df = {}
        log.debug('Building pandas dataframes:')
        for sample, rdf in self._d_rdf.items():
            log.debug(f'    {sample}')

            rep = rdf.Report()
            rep.Print()

            l_var      = [ name.c_str() for name in rdf.GetColumnNames() if self._need_var( name.c_str() )]
            data       = rdf.AsNumpy(l_var)
            df         = pnd.DataFrame(data)
            df['proc'] = sample

            d_df[sample] = df

        return d_df
    #-----------------------------------------------------------
    def _add_dec_weights(self, sample : str, df : pnd.DataFrame) -> pnd.DataFrame:
        if len(df) == 0:
            return df

        dec = self._d_wg['dec']

        if   dec == 1:
            log.debug(f'Adding decay weights to: {sample}')
            df['wgt_dec'] = df.apply(inclusive_decays_weights.read_weight, args=('L1', 'L2', 'H'), axis=1)
        elif dec == 0:
            log.warning(f'Not using decay weights in: {sample}')
            df['wgt_dec'] = 1.
        else:
            raise ValueError(f'Invalid value of wgt_dec: {dec}')

        arr_wgt      = df.wgt_dec.to_numpy()
        arr_wgt      = self._normalize_weights(arr_wgt)
        df['wgt_dec']= arr_wgt

        return df
    #-----------------------------------------------------------
    def _add_sam_weights(self, df : pnd.DataFrame) -> pnd.DataFrame:
        if len(df) == 0:
            return df

        sam = self._d_wg['sam']

        if   sam == 1:
            log.debug('Adding sample weights')
            obj           = inclusive_sample_weights(df)
            df['wgt_sam'] = obj.get_weights()
        elif sam == 0:
            log.warning('Not using sample weights')
            df['wgt_sam'] = 1.
        else:
            raise ValueError(f'Invalid value of wgt_sam: {sam}')

        arr_wgt      = df.wgt_sam.to_numpy()
        arr_wgt      = self._normalize_weights(arr_wgt)
        df['wgt_sam']= arr_wgt

        return df
    #-----------------------------------------------------------
    def _check_weights(self):
        try:
            [(k1, v1), (k2, v2)] = self._d_wg.items()
        except:
            log.error(f'Cannot extract two weight flags from: {self._d_wg}')
            raise

        if ([k1, k2] != ['dec', 'sam'])  and ([k1, k2] != ['sam', 'dec']):
            raise ValueError(f'Invalid weight keys: {k1}, {k2}')

        if (v1 not in [0, 1]) or (v2 not in [0, 1]):
            raise ValueError(f'Invalid weight values: {v1}, {v2}')
    #-----------------------------------------------------------
    def _check_valid(self, var, l_var, name):
        if var not in l_var:
            log.error(f'Value for {name}, {var}, is not valid')
            raise ValueError
    #-----------------------------------------------------------
    def _get_match_str(self) -> dict[str,str]:
        '''
        Returns
        ----------------
        _selection_ needed to split the charmonium sample into components
        for plotting
        '''
        if   self._q2bin == 'jpsi':
            d_match = self._get_match_str_jpsi()
        elif self._q2bin == 'psi2':
            d_match = self._get_match_str_psi2()
        elif self._q2bin in ['low', 'central', 'high']:
            d_match = self._get_match_str_psi2()
        else:
            raise ValueError(f'Invalid q2bin: {self._q2bin}')

        return d_match
    #-----------------------------------------------------------
    def _get_match_str_jpsi(self) -> dict[str,str]:
        bd          = '(abs(B_TRUEID) == 511)'
        bp          = '(abs(B_TRUEID) == 521)'
        bs          = '(abs(B_TRUEID) == 531)'

        d_cut                                  = {}
        d_cut[r'$B_d\to c\bar{c}(\to ee)H_s$'] = bd
        d_cut[r'$B^+\to c\bar{c}(\to ee)H_s$'] = bp
        d_cut[r'$B_s\to c\bar{c}(\to ee)H_s$'] = bs

        return d_cut
    #-----------------------------------------------------------
    def _get_match_str_psi2(self) -> dict[str,str]:
        bd          = '(abs(B_TRUEID) == 511)'
        bp_psjp     = '(abs(B_TRUEID) == 521) & (abs(Jpsi_TRUEID) == 443) & (abs(Jpsi_MC_MOTHER_ID) == 100443) & (abs(Jpsi_MC_GD_MOTHER_ID) == 521) & (abs(H_MC_MOTHER_ID) == 521)'
        bs          = '(abs(B_TRUEID) == 531)'

        neg_bp_psjp = bp_psjp.replace('==', '!=').replace('&' , '|')
        bp_ex       = f'(abs(B_TRUEID) == 521) & ({neg_bp_psjp})'

        d_cut       = {}
        d_cut[r'$B^+\to \psi(2S)(\to J/\psi+X)H_{s}$'] = bp_psjp
        d_cut[r'$B^+\to c\bar{c}(\to ee)H_s$']         = bp_ex
        d_cut[r'$B_d\to c\bar{c}(\to ee)H_s$']         = bd
        d_cut[r'$B_s\to c\bar{c}(\to ee)H_s$']         = bs

        return d_cut
    #-----------------------------------------------------------
    def _get_match_str_psi2_large(self) -> dict[str,str]:
        '''
        Returns dictionary needed to split mix of MC inclusive samples
        '''
        # pylint: disable=too-many-locals

        bp_psjp     = '(abs(Jpsi_MC_MOTHER_ID) == 100443) & (abs(Jpsi_MC_GD_MOTHER_ID) == 521) & (abs(H_MC_MOTHER_ID) == 521)'
        bd_psks     = '(abs(Jpsi_MC_MOTHER_ID) ==    511) & (abs(H_MC_MOTHER_ID) == 313) & (abs(H_MC_GD_MOTHER_ID) == 511) & (abs(Jpsi_TRUEID) == 100443)'
        bp_psks     = '(abs(Jpsi_MC_MOTHER_ID) ==    521) & (abs(H_MC_MOTHER_ID) == 323) & (abs(H_MC_GD_MOTHER_ID) == 521) & (abs(Jpsi_TRUEID) == 100443)'

        neg_bp_psjp = bp_psjp.replace('==', '!=').replace('&' , '|')
        neg_bd_psks = bd_psks.replace('==', '!=').replace('&' , '|')
        neg_bp_psks = bp_psks.replace('==', '!=').replace('&' , '|')

        bp_jpkp     = '(abs(B_TRUEID) == 521) & (abs(H_TRUEID) == 321) & (abs(Jpsi_TRUEID) == 443)'
        bd_jpkp     = '(abs(B_TRUEID) == 511) & (abs(H_TRUEID) == 321) & (abs(Jpsi_TRUEID) == 443)'

        bp_jpkp_ex  = f'({bp_jpkp}) & ({neg_bp_psjp}) & ({neg_bd_psks}) & ({neg_bp_psks})'
        bd_jpkp_ex  = f'({bd_jpkp}) & ({neg_bp_psjp}) & ({neg_bd_psks}) & ({neg_bp_psks})'

        neg_bp_jpkp = bp_jpkp.replace('==', '!=').replace('&' , '|')
        neg_bd_jpkp = bd_jpkp.replace('==', '!=').replace('&' , '|')


        bs          = '(abs(B_TRUEID) == 531)'
        neg_bs      = '(abs(B_TRUEID) != 531)'

        none        = f'({neg_bp_jpkp}) & ({neg_bd_jpkp}) & ({neg_bp_psjp}) & ({neg_bd_psks}) & ({neg_bp_psks}) & ({neg_bs})'

        d_cut            = {}
        d_cut['bp_psjp'] = bp_psjp
        d_cut['bp_psks'] = bp_psks
        d_cut['bp_jpkp'] = bp_jpkp_ex

        d_cut['bd_psks'] = bd_psks
        d_cut['bd_jpkp'] = bd_jpkp_ex

        d_cut['bs']      = bs

        d_cut['unmatched'] = none

        return d_cut
    #-----------------------------------------------------------
    def _get_match_str_psi2_all(self) -> dict[str,str]:
        d_cut           = {}
        d_cut['jpsi']   = '(Jpsi_TRUEID == 443)'
        d_cut['nojpsi'] = '(Jpsi_TRUEID != 443)'

        return d_cut
    #-----------------------------------------------------------
    def _print_wgt_stat(self, arr_wgt):
        l_wgt = arr_wgt.tolist()
        s_wgt = set(l_wgt)

        log.debug('-' * 20)
        log.debug(f'{"Frequency":<10}{"Weight":>10}')
        for wgt in s_wgt:
            nwgt = numpy.count_nonzero(wgt == arr_wgt)
            log.debug(f'{nwgt:<10}{wgt:>10.3}')
    #-----------------------------------------------------------
    def _normalize_weights(self, arr_wgt):
        tot_wgt = arr_wgt.sum()
        num_wgt = arr_wgt.shape[0]
        fact    = num_wgt / tot_wgt
        arr_wgt = fact * arr_wgt

        return arr_wgt
    #-----------------------------------------------------------
    def _filter_mass(self, df : pnd.DataFrame, mass : str, obs):
        ([[minx]], [[maxx]]) = obs.limits

        cut   = f'({minx} < {mass}) & ({mass} < {maxx})'
        log.debug(f'Applying: {cut}')
        inum  = df.shape[0]
        df    = df.query(cut)
        fnum  = df.shape[0]

        self._d_fstat[cut] = inum, fnum

        return df
    #-----------------------------------------------------------
    def _filter_cut(self, cut : str) -> pnd.DataFrame:
        if cut is None:
            log.debug('Not applying any cut')
            return self._df

        log.info(f'Applying cut: {cut}')
        inum = self._df.shape[0]
        df   = self._df.query(cut)
        fnum = df.shape[0]

        self._d_fstat[cut] = inum, fnum

        return df
    #-----------------------------------------------------------
    def _get_identifier(self, mass : str, cut : str, **kwargs) -> str:
        cwargs = copy.deepcopy(kwargs)
        del cwargs['obs']

        swgt = json.dumps(self._d_wg , sort_keys=True)
        scwg = json.dumps(cwargs     , sort_keys=True)

        l_d_sel   = [ sel.selection(trigger=self._trig, q2bin=self._q2bin, process=sample) for sample in self._l_sample ]
        l_element = [
                swgt,
                self._trig,
                self._q2bin,
                mass,
                scwg,           # Stringified keyword arguments
                self._l_sample, # ccbar cocktail sample names
                l_d_sel,        # list of selections, one for each ccbar cocktail sample
                cut]

        hsh  = hashing.hash_object(l_element)

        return hsh
    #-----------------------------------------------------------
    def _path_from_identifier(self, identifier : str) -> str:
        dir_path = '/tmp/cache/prec'
        os.makedirs(dir_path, exist_ok=True)

        return f'{dir_path}/pdf_{identifier}.json'
    #-----------------------------------------------------------
    def _drop_before_saving(self, df : pnd.DataFrame) -> pnd.DataFrame:
        l_needed = self._l_mass + ['wgt_br', 'wgt_dec', 'wgt_sam']
        l_drop   = [ name for name in df.columns if  name not in l_needed ]
        df       = df.drop(l_drop, axis=1)

        return df
    #-----------------------------------------------------------
    def _get_pdf(
        self,
        mass           : str,
        df             : pnd.DataFrame,
        component_name : str,
        **kwargs) -> zpdf|None:
        '''
        Parameters
        ------------------
        name    : Latex name of PDF component
        mass    : Mass, with values in:
            mass     : Non constrained B mass
            mass_jpsi: Jpsi constrained B mass
            mass_psi2: Psi2S constrained B mass
        df      : DataCorresponding to given ccbar component
        **kwargs: These are all arguments for KDE1DimISJ or KDE1DimFFT

        Returns
        ------------------
        Either:
        None   : If there are fewer than _min_entries
        KDE PDF: Otherwise
        '''
        # This kwargs reffers to this particular PDF
        kwargs         = copy.deepcopy(kwargs)
        pdf_name       = kwargs['name']
        slug           = slugify.slugify(pdf_name, lowercase=False)
        kwargs['name'] = component_name

        arr_mass = df[mass     ].to_numpy()
        arr_wgt  = df['wgt_br' ].to_numpy()
        obs      = kwargs['obs']

        nentries = self._yield_from_arrays(arr_mass=arr_mass, arr_weight=arr_wgt, obs=obs)
        if nentries < self._min_entries:
            log.warning(f'Found fewer than {self._min_entries}: {nentries}, skipping PDF {component_name}')
            return None

        log.debug(f'Building PDF with {nentries} entries for {component_name}')

        pdf          = self._pdf_from_df(df=df, mass=mass, **kwargs)
        pdf.arr_mass = arr_mass
        pdf.arr_wgt  = arr_wgt
        pdf.arr_sam  = df['wgt_sam'].to_numpy()
        pdf.arr_dec  = df['wgt_dec'].to_numpy()

        if not is_pdf_usable(pdf):
            log.warning(f'PDF {component_name} is not usable')
            return None

        PRec.plot_pdf(
            pdf,
            title  =component_name,
            name   =component_name,
            out_dir=f'{self._out_path}/{slug}')

        return pdf
    #-----------------------------------------------------------
    def _pdf_from_df(
        self,
        df   : pnd.DataFrame,
        mass : str,
        **kwargs) -> zpdf:
        '''
        Will build KDE from dataframe with information needed

        Parameters
        ---------------
        df     : DataFrame with weights and masses, the weight is assumed to be in 'wgt_br'
        mass   : Name of the column with mass to be fitted
        kwargs : Keyword arguments meant to be passed to KDE1Dim*
        '''
        arr_mass = df[mass    ].to_numpy()
        arr_wgt  = df['wgt_br'].to_numpy()
        arr_wgt  = arr_wgt.astype(float)

        try:
            pdf = self._build_kde(arr_mass=arr_mass, arr_wgt=arr_wgt, **kwargs)
        except Exception as exc:
            for setting, value in kwargs.items():
                if not isinstance(value, (str,float,int)):
                    log.info(f'{setting:<30}{"--->"}')
                    log.info(value)
                else:
                    log.info(f'{setting:<30}{value:<30}')

            raise Exception('Failed to build KDE') from exc

        return pdf
    #-----------------------------------------------------------
    def _build_kde(
            self,
            arr_mass : numpy.ndarray,
            arr_wgt  : numpy.ndarray, **kwargs) -> zpdf:
        '''
        Parameters
        --------------
        arr_xxx: Contains mass and weights needed for KDE

        Returns
        --------------
        Either FFT or ISJ KDE
        '''
        if log.getEffectiveLevel() < 20:
            log.debug('Using fitting options:')
            pprint.pprint(kwargs)

        nentries = len(arr_mass)
        if nentries > self._min_isj_entries:
            log.debug('Using ISJ KDE for high statistics sample')
            if 'bandwidth' in kwargs: # ISJ does not accept this argument
                del kwargs['bandwidth']

            pdf = zfit.pdf.KDE1DimISJ(arr_mass, weights=arr_wgt, **kwargs)
        else:
            log.debug('Using FFT KDE for low statistics sample')
            pdf = zfit.pdf.KDE1DimFFT(arr_mass, weights=arr_wgt, **kwargs)

        return pdf
    #-----------------------------------------------------------
    def _print_cutflow(self) -> None:
        log.debug('-' * 50)
        log.debug(f'{"Cut":<30}{"Total":<20}{"Passed":<20}')
        log.debug('-' * 50)
        for cut, (inum, fnum) in self._d_fstat.items():
            log.debug(f'{cut:<30}{inum:<20}{fnum:<20}')
        log.debug('-' * 50)
    #-----------------------------------------------------------
    def _frac_from_pdf(self, pdf : zpdf, frc : float) -> zpar:
        name = pdf.name
        slug = slugify.slugify(name, lowercase=False)
        par  = zfit.param.Parameter(f'f_{slug}', frc, 0, 1)

        par.floating = False

        return par
    #-----------------------------------------------------------
    def _yield_in_range(self, pdf : zpdf) -> float:
        '''
        Parameters
        ---------------
        pdf: ZFit KDE PDF with array of weights and masses that were used to make it, attached

        Returns
        ---------------
        The mass and weights are defined in the WHOLE range. This method extracts the yields
        in the observable range. Needed to calculate fractions of componets, used to put
        ccbar stuff together
        '''
        obs = pdf.space
        wgt = pdf.arr_wgt
        mas = pdf.arr_mass

        return self._yield_from_arrays(arr_mass=mas, arr_weight=wgt, obs=obs)
    #-----------------------------------------------------------
    def _yield_from_arrays(
        self,
        obs        : zobs,
        arr_mass   : numpy.ndarray,
        arr_weight : numpy.ndarray) -> float:

        minx, maxx = sut.range_from_obs(obs=obs)

        mask= (minx < arr_mass) & (arr_mass < maxx)
        wgt = arr_weight[mask]

        return sum(wgt)
    #-----------------------------------------------------------
    def _get_full_pdf(
        self,
        mass : str,
        d_df : dict[str,pnd.DataFrame],
        **kwargs) -> zpdf|None:
        '''
        Parameters
        -------------------
        mass  : Name of the column in the dataframe, which will be used to build KDE
        kwars : Key word arguments needed to build KDE PDF
        d_df  : Dictionary with:
            Key  : Latex name of component, not necessarily Bu/Bd... This was re-split
            Value: Dataframe with data to fit

        Returns
        -------------------
        Full pdf, i.e. all ccbar components added
        '''
        l_pdf     = [ self._get_pdf(mass = mass, component_name = ltex, df = df, **kwargs) for ltex, df in d_df.items()                    ]
        l_pdf     = [                                                                  pdf for      pdf in l_pdf        if pdf is not None ]
        l_wgt_yld = [ self._yield_in_range(pdf=pdf)                                        for      pdf in l_pdf                           ]
        l_frc     = [ wgt_yld / sum(l_wgt_yld)                                             for  wgt_yld in l_wgt_yld                       ]
        l_yld     = [ self._frac_from_pdf(pdf=pdf, frc=frc)                                for pdf, frc in zip(l_pdf, l_frc)               ]

        if   len(l_pdf) >= 2:
            pdf   = zfit.pdf.SumPDF(l_pdf, fracs=l_yld, name='ccbar PRec')
        elif len(l_pdf) == 1:
            [pdf] = l_pdf
        else:
            log.warning('No PDF can be built with dataset')
            return None

        l_arr_mass   = [ pdf.arr_mass for pdf in l_pdf ]
        l_arr_wgt    = [ pdf.arr_wgt  for pdf in l_pdf ]
        l_arr_sam    = [ pdf.arr_sam  for pdf in l_pdf ]
        l_arr_dec    = [ pdf.arr_dec  for pdf in l_pdf ]

        pdf.arr_mass = numpy.concatenate(l_arr_mass)
        pdf.arr_wgt  = numpy.concatenate(l_arr_wgt )
        pdf.arr_dec  = numpy.concatenate(l_arr_dec )
        pdf.arr_sam  = numpy.concatenate(l_arr_sam )

        return pdf
    #-----------------------------------------------------------
    def get_sum(self, mass : str, name='unnamed', **kwargs) -> Union[zpdf,None]:
        '''Provides extended PDF that is the sum of multiple KDEs representing PRec background

        Parameters:
        mass (str) : Defines which mass constrain to use, choose between "B_M", "B_const_mass_M", "B_const_mass_psi2S_M"
        name (str) : PDF name
        **kwargs   : Arguments meant to be taken by zfit KDE1DimFFT

        Returns:
        zfit.pdf.SumPDF instance
        '''
        kwargs['name'] = name
        slug           = slugify.slugify(name, lowercase=False)

        l_ltex      = list(self._d_match) # Get component names in latex and map them to parquet files to save
        d_ltex_slug = { ltex : slugify.slugify(ltex, lowercase=False) for ltex       in l_ltex }
        d_path      = { ltex : f'{self._out_path}/{slug}.parquet'     for ltex, slug in d_ltex_slug.items() }

        if self._copy_from_cache():
            log.info(f'Data found cached, reloading from {self._out_path}')
            d_df = { ltex : pnd.read_parquet(path) for ltex , path in d_path.items() }
            pdf        = self._get_full_pdf(mass=mass, d_df=d_df, **kwargs)

            PRec.plot_pdf(
                pdf,
                title  =name,
                name   =name,
                out_dir=f'{self._out_path}/{slug}')

            return pdf

        log.info(f'Recalculating, cached data not found in: {self._out_path}')

        d_df = self._get_df()
        pdf  = self._get_full_pdf(mass=mass, d_df=d_df, **kwargs)

        # Save dataframes before caching
        # Cache at the very end
        log.info('Saving dataframes:')
        for ltex, df in d_df.items():
            path = d_path[ltex]
            log.info(f'   {path}')
            df.to_parquet(path)

        PRec.plot_pdf(
            pdf,
            title  =name,
            name   =name,
            out_dir=f'{self._out_path}/{slug}')

        self._cache()

        return pdf
    #-----------------------------------------------------------
    @staticmethod
    def plot_pdf(
            pdf     : zpdf|None,
            name    : str,
            out_dir : str,
            title   : str        = '',
            maxy    : float|None = None) -> None:
        '''
        Utility method, meant to plot PDF after it was built

        Parameters
        ------------------
        pdf    : PDF
        neme   : used to name the PNG file as {name}.png
        title  : Title for plots, will be appended after number of entries
        maxy   : Will be used to plot fit properly in case labels overlap
        out_dir: Directory where plots will go
        '''

        if pdf is None:
            log.warning(f'PDF {name} not build, not plotting')
            return

        arr_mass = pdf.arr_mass
        arr_wgt  = pdf.arr_wgt
        arr_sam  = pdf.arr_sam
        arr_dec  = pdf.arr_dec

        obj = ZFitPlotter(data=arr_mass, model=pdf, weights=arr_wgt)
        obj.plot(stacked=True)

        obj.axs[0].set_title(f'#Entries: {arr_mass.size}; {title}')

        if maxy is not None:
            obj.axs[0].set_ylim(bottom=0, top=maxy)

        obj.axs[0].axvline(x=5080, linestyle=':')
        obj.axs[0].axvline(x=5680, linestyle=':')
        obj.axs[0].axvline(x=5280, label=r'$B^+$', color='gray', linestyle='--')

        obj.axs[1].set_ylim(-5, +5)
        obj.axs[1].axhline(y=-3, color='red')
        obj.axs[1].axhline(y=+3, color='red')
        obj.axs[1].set_label('M$(B^+)$[MeV/${}_{c^2}$]')

        os.makedirs(out_dir, exist_ok=True)

        slug = slugify.slugify(name, lowercase=False)

        plot_path = f'{out_dir}/{slug}.png'
        log.info(f'Saving to: {plot_path}')
        plt.savefig(plot_path)
        plt.close('all')

        plt.hist(arr_sam, bins=30, label='sample', histtype='step', linestyle='-' )
        plt.hist(arr_dec, bins=30, label='decay' , histtype='step', linestyle='--')
        plt.hist(arr_wgt, bins=30, label='Total' , histtype='step', linestyle=':' )

        plt.legend()
        plt.title(title)
        plt.savefig(f'{out_dir}/{slug}_wgt.png')
        plt.close('all')

        text_path = plot_path.replace('png', 'txt')
        sut.print_pdf(pdf, txt_path=text_path)
    #-----------------------------------------------------------
    @staticmethod
    @contextmanager
    def apply_setting(use_cache : bool):
        '''
        Used to override default behaviour

        use_cache : If False (default is True) will recalculate the PDF
        '''
        old_val = PRec.use_cache
        try:
            PRec.use_cache = use_cache
            yield
        finally:
            PRec.use_cache = old_val
#-----------------------------------------------------------



'''
Module holding DataPreprocessor class
'''
from typing import cast

import numpy

from ROOT                   import RDataFrame
from dmu.workflow.cache     import Cache
from dmu.stats.zfit         import zfit
from dmu.stats              import utilities  as sut
from dmu.logging.log_store  import LogStore
from zfit.core.interfaces   import ZfitData   as zdata
from zfit.core.interfaces   import ZfitSpace  as zobs
from rx_data.rdf_getter     import RDFGetter
from rx_selection           import selection  as sel

log=LogStore.add_logger('fitter:data_preprocessor')
# ------------------------
class DataPreprocessor(Cache):
    '''
    Class in charge of providing datasets for fitting by:

    - Loading ROOT files through RDFGetter
    - Applying selection
    - Transforming data into format that zfit can use
    '''
    # ------------------------
    def __init__(
            self,
            out_dir : str,
            obs     : zobs,
            sample  : str,
            trigger : str,
            project : str,
            q2bin   : str,
            cut     : str|None = None):
        '''
        Parameters
        --------------------
        out_dir: Directory where caching will happen, with respect to the _cache_root directory
        obs    : zfit observable
        sample : e.g. DATA_24_MagUp...
        trigger: e.g. Hlt2RD...
        project: e.g. rx, nopid
        q2bin  : e.g. central
        cut    : selection that can be added on top. Needed when fits are required in categories, optional
        '''
        self._obs    = obs
        self._sample = sample
        self._trigger= trigger
        self._project= project
        self._q2bin  = q2bin
        self._rdf    = self._get_rdf(cut=cut)
        self._rdf_uid= None if self._rdf is None else self._rdf.uid

        super().__init__(
            out_path = out_dir,
            obs_name = sut.name_from_obs(obs),
            rdf_uid  = self._rdf_uid)
    # ------------------------
    def _get_rdf(self, cut : str|None) -> RDataFrame|None:
        '''
        Parameters
        -------------------
        category_cut: Selection to be added on top, used for categories

        Returns
        -------------------
        Either:

        - If dataset is not a toy, ROOT dataframe after selection and with Unique identifier attached as uid
        - Otherwise, None
        '''
        if 'toy' in self._sample:
            log.debug(f'Cannot retrieve dataframe for toy sample: {self._sample}')
            return None

        log.debug('Retrieving dataframe')
        gtr = RDFGetter(sample =self._sample, trigger=self._trigger)
        rdf = gtr.get_rdf()
        uid = gtr.get_uid()
        rdf = cast(RDataFrame, rdf)

        log.debug('Applying selection')
        rdf = sel.apply_full_selection(
            rdf     = rdf,
            uid     = uid,
            q2bin   = self._q2bin,
            trigger = self._trigger,
            process = self._sample,
            ext_cut = cut)

        return rdf
    # ------------------------
    def _get_toy_array(self, sample : str) -> numpy.ndarray:
        '''
        Returns array with toy data
        '''
        if sample == 'gauss_toy':
            sig  = numpy.random.normal(loc=5280, scale=50, size=10_000)
            return sig

        if sample == 'data_toy':
            sig = self._get_toy_array(sample='gauss_toy')
            bkg = numpy.random.exponential(scale=10_000, size=100_000)
            arr = numpy.concatenate((sig, bkg))

            return arr

        raise NotImplementedError(f'Cannot retrive toy data for sample: {sample}')
    # ------------------------
    def _get_array(self) -> tuple[numpy.ndarray,numpy.ndarray]:
        '''
        Return a tuple of numpy arrays with the observable and weight
        for the sample requested, this array is fully selected
        '''
        if 'toy' in self._sample:
            log.debug(f'Extracting toy data for sample {self._sample}')
            arr = self._get_toy_array(sample=self._sample)
            wgt = numpy.ones_like(arr)

            return arr, wgt

        log.debug(f'Extracting data through RDFGetter for sample {self._sample}')

        rdf = self._rdf
        if log.getEffectiveLevel() < 20:
            rep = rdf.Report()
            rep.Print()

        name = sut.name_from_obs(obs=self._obs)

        log.debug('Retrieving data')
        arr  = rdf.AsNumpy([name])[name]
        wgt  = rdf.AsNumpy(['weight'])['weight']

        nevt = len(arr)
        log.debug(f'Found {nevt} entries')

        return arr, wgt
    # ------------------------
    @property
    def rdf_uid(self) -> str|None:
        '''
        Unique identifier of ROOT dataframe after selection
        '''
        return self._rdf_uid
    # ------------------------
    def _data_from_numpy(
            self,
            arr_value : numpy.ndarray,
            arr_weight: numpy.ndarray) -> zdata:
        '''
        We should not use weights if they are all 1s due to problems in zfit

        Parameters
        ------------
        arr_value : Array with values to be fitted
        arr_weight: Array with weights

        Returns
        ------------
        zfit data.
        '''

        arr_is_close = numpy.isclose(arr_weight, 1.0, rtol=1e-5)

        if numpy.all(arr_is_close):
            log.debug('Not using weights for dataset where all weights are 1')
            wgt = None
        else:
            log.debug('Using weights in dataset')
            wgt = arr_weight

        data = zfit.data.from_numpy(obs=self._obs, array=arr_value, weights=wgt)

        return data
    # ------------------------
    def get_data(self) -> zdata:
        '''
        Returns zfit data
        '''
        data_path = f'{self._out_path}/data.npz'
        if self._copy_from_cache():
            log.warning(f'Data found cached, loading: {data_path}')
            with numpy.load(data_path) as ifile:
                arr = ifile['values' ]
                wgt = ifile['weights']

            data    = self._data_from_numpy(arr_value=arr, arr_weight=wgt)
            return data

        arr, wgt = self._get_array()
        data     = self._data_from_numpy(arr_value=arr, arr_weight=wgt)

        numpy.savez_compressed(data_path, values=arr, weights=wgt)
        self._cache()

        return data
# ------------------------



'''
Module with functions intended to interface with the PDG API
'''

import pdg

from dmu.logging.log_store import LogStore

log=LogStore.add_logger('fitter:pdg_utils')

# TODO: We should not connect to the PDF to get a few branching fractions
# Put the BF in a YAML file and load it from there
#-------------------------------------------------------
def get_bf(decay : str) -> float:
    '''
    Returns branching fraction for a given decay
    '''
    api    = pdg.connect()
    mother = decay.split('-->')[0].replace(' ', '')
    for bf in api.get_particle_by_name(mother).exclusive_branching_fractions():
        if bf.is_limit:
            continue

        if bf.description == decay:
            return bf.value

    log.error(f'Cannot find BF for decay: {decay}')
    raise ValueError('Make sure your pdg>=0.1.2')
#-------------------------------------------------------



'''
Module with SimFitter class
'''

from typing import cast

from omegaconf                import DictConfig, OmegaConf

from dmu.stats.zfit           import zfit
from dmu.stats                import utilities    as sut
from dmu.workflow.cache       import Cache
from dmu.stats.model_factory  import ModelFactory
from dmu.logging.log_store    import LogStore

from rx_efficiencies.decay_names import DecayNames
from rx_selection             import selection    as sel
from zfit.core.interfaces     import ZfitData     as zdata
from zfit.core.interfaces     import ZfitPDF      as zpdf
from zfit.core.parameter      import Parameter    as zpar
from zfit.result              import FitResult    as zres
from zfit.core.interfaces     import ZfitSpace    as zobs
from fitter.base_fitter       import BaseFitter
from fitter.data_preprocessor import DataPreprocessor
from fitter.prec              import PRec

log=LogStore.add_logger('fitter:sim_fitter')
# ------------------------
class SimFitter(BaseFitter, Cache):
    '''
    Fitter for simulation samples
    '''
    # ------------------------
    def __init__(
        self,
        component : str,
        trigger : str,
        project : str,
        q2bin   : str,
        cfg     : DictConfig,
        obs     : zobs,
        name    : str|None = None):
        '''
        Parameters
        --------------------
        obs      : Observable
        name     : Optional, identifier for fit, used to name directory
        component: Nickname of component, e.g. combinatorial, only used for naming
        trigger  : Hlt2RD...
        project  : E.g. rx
        q2bin    : E.g. central
        cfg      : Object storing configuration for fit
        '''
        BaseFitter.__init__(self)

        self._sample    = DecayNames.sample_from_decay(component, fall_back='NA')
        self._name      = name
        self._component = component
        self._trigger   = trigger
        self._project   = project
        self._q2bin     = q2bin
        self._cfg       = cfg
        self._obs       = obs
        if name is None:
            self._base_path = f'{cfg.output_directory}/{trigger}_{project}_{q2bin}'
        else:
            self._base_path = f'{cfg.output_directory}/{name}/{trigger}_{project}_{q2bin}'

        log.debug(f'For component {self._component} using output: {self._base_path}')

        self._l_rdf_uid = []
        self._d_data    = self._get_data()

        # Will not build KDE if fewer than these entries in dataset
        self._min_kde_entries = 50

        # Will not build (fit) a parametric PDF if fewer than these entries
        # will return None
        self._min_fit_entries = 100

        # All the PDFs will share the mu and sigma below and these will float
        self._mu_par = zfit.param.Parameter('mu_flt', 5280, 5000, 5500)
        self._sg_par = zfit.param.Parameter('sg_flt',   15,    5,  300)

        Cache.__init__(
            self,
            out_path = self._base_path,
            l_rdf_uid= self._l_rdf_uid,
            config   = OmegaConf.to_container(cfg, resolve=True))
    # ------------------------
    def _get_data(self) -> dict[str,zdata]:
        '''
        Returns
        --------------------
        dictionary with:

        Key  : Name of MC category, e.g. brem category
        Value: Zfit dataset
        '''
        d_data = {}
        # For components without an MC associated e.g. combinatorial
        # return empty dataset
        if 'sample' not in self._cfg:
            return d_data

        for cat_name, data in self._cfg.categories.items():
            cat_cut = None if 'selection' not in data else data.selection

            log.debug(f'Using category {cat_name} with cut {cat_cut}')

            prp   = DataPreprocessor(
                obs    = self._obs,
                cut    = cat_cut,
                trigger= self._trigger,
                project= self._project,
                q2bin  = self._q2bin,
                out_dir= self._base_path,
                sample = self._cfg.sample)
            d_data[cat_name] = prp.get_data()

            self._l_rdf_uid.append(prp.rdf_uid)

        return d_data
    # ------------------------
    def _get_pdf(
            self,
            cfg     : DictConfig,
            category: str,
            l_model : list[str]) -> zpdf:
        '''
        Parameters
        ------------
        category: If the MC is meant to be split (e.g. by brem) this should the the label of the category
        cfg     : DictConfig with model configuration, stores parameters that are floating, fixed, etc
        l_model : List of model names, e.g. [cbl, cbr]

        Returns
        ------------
        Fitting PDF built from the sum of those models
        '''
        log.info(f'Building {self._component} for category {category} with: {l_model}')

        mod     = ModelFactory(
            preffix = f'{self._component}_{category}',
            obs     = self._obs,
            l_pdf   = l_model,
            l_reuse = [self._mu_par, self._sg_par],
            l_shared= cfg.shared,
            l_float = cfg.float ,
            d_rep   = cfg.reparametrize,
            d_fix   = cfg.fix)

        pdf = mod.get_pdf()

        return pdf
    # ------------------------
    def _fix_tails(self, pdf : zpdf, pars : DictConfig) -> zpdf:
        '''
        Parameters
        --------------
        pdf : PDF after fit
        pars:
        '''
        s_par = pdf.get_params()
        npar  = len(s_par)
        log.debug(f'Found {npar} floating parameters')

        for par in s_par:
            # Model builder adds _flt to name
            # of parameters meant to float
            if par.name.endswith('_flt'):
                log.debug(f'Not fixing: {par.name}')
                continue

            if par.name in pars:
                par.set_value(pars[par.name].value)
                log.debug(f'{par.name:<20}{"--->"}{pars[par.name].value:>20.3f}')
                par.floating = False

        return pdf
    # ------------------------
    def _get_nomc_component(self) -> zpdf:
        '''
        This method will return a PDF when there is no simulation
        associated to it, e.g. Combinatorial
        '''
        if 'main' not in self._cfg.categories:
            log.info(OmegaConf.to_yaml(self._cfg))
            raise ValueError(f'Cannot find main category in config associated to sample {self._component}')

        l_model = self._cfg.categories.main.models[self._q2bin]
        cfg     = self._cfg[self._q2bin]

        model   = self._get_pdf(
            l_model = l_model,
            cfg     = cfg,
            category= 'main')

        return model
    # ------------------------
    def _fit_category(
            self,
            skip_fit     : bool,
            category     : str,
            l_model_name : list[str]) -> tuple[zpdf|None,float|None,zres|None]:
        '''
        Parameters
        ----------------
        skip_fit     : If true, it will only return model, used if fit parameters were already found
        category     : Name of fitting category
        l_model_name : List of fitting models,  e.g. [cbr, cbl]

        Returns
        ----------------
        Tuple with:
            - Fitted PDF, None if problems were found building it, e.g. too few entries
            - Size (sum of weights) of dataset in given category.
              If fit is skipped, returns None, because this is used to set
              the value of the fit fraction, which should already be in the cached data.
            - zfit result object, if fit is skipped, returns None
        '''
        log.info(f'Fitting category {category}')

        model = self._get_pdf(
            category= category,
            cfg     = self._cfg,
            l_model = l_model_name)

        data  = self._d_data[category]

        sumw  = sut.yield_from_zdata(data=data)
        if skip_fit:
            return model, sumw, None

        if sumw < self._min_fit_entries:
            log.warning(f'Found to few entries {sumw:.1f} < {self._min_fit_entries}, skipping component')
            return None, 0, None

        res   = self._fit(data=data, model=model, cfg=self._cfg.fit)
        self._save_fit(
            cuts     = sel.selection(process=self._cfg.sample, trigger=self._trigger, q2bin=self._q2bin),
            cfg      = self._cfg.plots,
            data     = data,
            model    = model,
            res      = res,
            out_path = f'{self._out_path}/{category}')

        cres  = sut.zres_to_cres(res=res)
        model = self._fix_tails(pdf=model, pars=cres)

        return model, sumw, res
    # ------------------------
    # TODO: Fractions need to be parameters to be constrained
    def _get_fraction(
        self,
        sumw     : float,
        total    : float,
        category : str) -> zpar:
        '''
        Parameters
        -------------
        sumw    : Yield in MC associated to this category
        total   : Total yield
        category: Name of this category

        Returns
        -------------
        Fitting fraction parameter fixed
        '''
        frac_name = f'frac_{self._component}_{category}'
        value     = sumw / total
        par       = zfit.param.Parameter(frac_name, value, 0, 1)

        log.debug(f'{frac_name:<50}{value:<10.3f}')

        return par
    # ------------------------
    def _get_full_model(self, skip_fit : bool) -> tuple[zpdf,DictConfig]:
        '''
        Parameters
        ---------------
        skip_fit: If true, it will rturn the model without fitting

        Returns
        ---------------
        Tuple with:

        - PDF for the combined categories with the parameters set
        to the fitted values
        - Instance of DictConfig storing all the fitting parameters
        '''
        l_pdf   = []
        l_yield = []
        l_cres  = []
        for category, data in self._cfg.categories.items():
            l_model_name     = data['model']
            model, sumw, res = self._fit_category(
                skip_fit     = skip_fit,
                category     = category,
                l_model_name = l_model_name)

            if model is None:
                log.warning(f'Skipping cateogory {category}')
                continue

            # Will be None if fit is cached
            # and this is only returning model
            cres = OmegaConf.create({})
            if res is not None:
                cres = sut.zres_to_cres(res)

            l_pdf.append(model)
            l_yield.append(sumw)
            l_cres.append(cres)

        return self._merge_categories(
            l_pdf  =l_pdf,
            l_yield=l_yield,
            l_cres =l_cres)
    # ------------------------
    def _merge_categories(
        self,
        l_pdf   : list[zpdf],
        l_yield : list[float],
        l_cres  : list[DictConfig]) -> tuple[zpdf,DictConfig]:
        '''
        Parameters
        -----------------
        l_pdf  : List of zfit PDFs from fit, one per category
        l_yield: List of yields from MC sample, not the fitted one
        l_cres : List of result objects holding parameter values from fits

        Returns
        -----------------
        Tuple with:

        - Full PDF, i.e. sum of components
        - Merged dictionary of parameters
        '''

        if len(l_pdf) == 1:
            cres  = OmegaConf.merge(l_cres[0])

            return l_pdf[0], cres

        log.debug(60 * '-')
        log.debug(f'{"Fraction":<50}{"Value":<10}')
        log.debug(60 * '-')
        l_frac = [
                  self._get_fraction(
                      sumw,
                      total   = sum(l_yield),
                      category= category)
                  for sumw, category in zip(l_yield, self._cfg.categories) ]
        log.debug(60 * '-')

        full_model = zfit.pdf.SumPDF(l_pdf, l_frac)
        full_cres  = OmegaConf.merge(*l_cres)

        return full_model, full_cres
    # ------------------------
    def _is_kde(self) -> bool:
        '''
        Returns true if the PDF is meant to be a single KDE
        False if it is meant to be a parametric PDF
        '''
        if len(self._cfg.categories) > 1:
            return False

        # By convention, if there is a single category
        # It will be the main category
        model_name = self._cfg.categories.main.model

        if not isinstance(model_name, str):
            return False

        if model_name.startswith('KDE'):
            return True

        raise ValueError(f'Invalid PDF found: {model_name}')
    # ------------------------
    def _get_kde(self) -> zpdf|None:
        '''
        - Makes KDE PDF
        - Saves fit (plot, list of parameters, etc)

        Returns
        ------------------
        - KDE PDF after fit
        - None if there are fewer than _min_kde_entries
        '''
        model_name = self._cfg.categories.main.model
        data       = self._d_data['main']

        KdeBuilder = getattr(zfit.pdf, model_name)
        if data.n_events < self._min_kde_entries:
            pdf = None
        else:
            if 'options' in self._cfg.fit:
                cfg_fit= self._cfg.fit.get('options')
                kwargs = OmegaConf.to_container(cfg_fit, resolve=True)
            else:
                kwargs = {}

            pdf = KdeBuilder(obs=self._obs, data=data, name=self._component, **kwargs)

        self._save_fit(
            cuts     = sel.selection(process=self._cfg.sample, trigger=self._trigger, q2bin=self._q2bin),
            cfg      = self._cfg.plots,
            data     = data,
            model    = pdf,
            res      = None,
            out_path = f'{self._out_path}/main')

        return pdf
    # ------------------------
    def _get_ccbar_component(self) -> zpdf|None:
        '''
        This is an interace to the PRec class, which is in
        charge of building the KDE for ccbar sample

        Returns
        ----------------
        Either:

        - PDF with KDEs added corresponding to different groups of ccbar decays
        - None, if no data was found
        '''
        ftr=PRec(
            trig    = self._trigger,
            q2bin   = self._q2bin  ,
            d_weight= self._cfg.weights,
            out_dir = self._base_path,
            samples = self._cfg.ccbar_samples)

        obs_name = sut.name_from_obs(self._obs)

        kwargs   = OmegaConf.to_container(self._cfg.fitting, resolve=True)

        pdf =ftr.get_sum(
            mass   = obs_name,
            name   = r'$c\bar{c}+X$',
            obs    = self._obs,
            **kwargs)

        return pdf
    # ------------------------
    def get_model(self) -> zpdf|None:
        '''
        Returns
        ------------
        zfit PDF, not extended yet
        '''
        if 'ccbar_samples' in self._cfg:
            return self._get_ccbar_component()

        if 'sample' not in self._cfg:
            return self._get_nomc_component()

        result_path = f'{self._out_path}/parameters.yaml'
        if self._copy_from_cache():
            res      = OmegaConf.load(result_path)
            res      = cast(DictConfig, res)
            # If caching, need only model, second return value
            # Is an empty DictConfig, because no fit happened
            model, _ = self._get_full_model(skip_fit=True)
            model    = self._fix_tails(pdf=model, pars=res)

            return model

        log.info(f'Fitting, could not find cached parameters in {result_path}')

        if self._is_kde():
            return self._get_kde()

        full_model, cres = self._get_full_model(skip_fit=False)

        OmegaConf.save(cres, result_path)

        self._cache()
        return full_model
# ------------------------



'''
Module with Reader class
'''
# pylint: disable=too-many-instance-attributes, invalid-name

import pandas as pnd
from fitter.pchain         import PChain
from dmu.logging.log_store import LogStore

log = LogStore.add_logger('rx_fitter:inclusive_decays_weights')
#---------------------------
class Reader:
    '''
    Class used to attach decay weights to dataframe of inclusive decays
    '''
    def __init__(self, l1, l2, kp):
        self._l1_ch = l1
        self._l2_ch = l2
        self._kp_ch = kp

        self._weight= 1.0

        self._Jp_id = 443
        self._Ps_id = 100443
        self._Pi_id = 211
        self._Ph_id = 333
        self._Et_id = 221
        self._Ks_id = 313
        self._Kp_id = 321
        self._KS_id = 310

        self._Bd_id = 511
        self._Bu_id = 521
        self._Bs_id = 531

        #From https://gitlab.cern.ch/LHCb-RD/ewp-rkstz/-/blob/master/analysis/kernel/inc/ConstDef.hpp
        self._Bc         = 541
        self._Lb         = 5122
        self._L0         = 3122
        self._Kst_c      = 323
        self._K_1_1270_z = 10313
        self._K_1_1270_c = 10323
        self._K_2_1430_c = 325
        self._K_2_1430_z = 315
        self._K_1_1400_z = 20313
        self._K_1_1400_c = 20323
        self._K0         = 311
        self._KLong      = 130
        self._Pi0        = 111
        self._M          = 13
        self._E          = 11
        self._Tau        = 15
        self._P          = 2212
        self._N          = 2112
        self._Eta_prime  = 331
        self._Rho0       = 113
        self._Rho_c      = 213
        self._Omega      = 223
        self._D0         = 421
        self._Dp         = 411
        self._Ds         = 431
        self._Bst0       = 513
        self._Bst_plus   = 523
        self._Photon     = 22

        self._l_bid = [self._Bd_id, self._Bu_id, self._Bs_id]
    #---------------------------
    def _get_jpsi_wgt(self) -> float:
        weight = 1.0
        for bid in  self._l_bid:
            flg_1 = self._l1_ch.MatchUpstream( self._Ps_id, bid)
            flg_2 = self._l2_ch.MatchUpstream( self._Ps_id, bid)

            if not flg_1 and not flg_2:
                continue

            flg_3 = self._l1_ch.MatchMother(self._Jp_id)
            flg_4 = self._l2_ch.MatchMother(self._Jp_id)
            flg_5 = self._l1_ch.MatchMother(self._Ps_id)
            flg_6 = self._l2_ch.MatchMother(self._Ps_id)

            if   flg_3 or flg_4:
                #weight = 0.6254 / ( 1-0.1741) #0.75
                weight = 0.958
            elif flg_5 or flg_6:
                #weight = 1.3200 / 0.1741 #7.58
                weight = 0.771

        return weight
    #---------------------------
    def _get_brfrac_corr(self) -> float:
        weight= 1
        flg_1 = self._l1_ch.MatchUpstream(self._Ps_id, self._Bd_id)
        flg_2 = self._l2_ch.MatchUpstream(self._Ps_id, self._Bd_id)
        if flg_1 or flg_2:
            weight = 1.17

        flg_1 = self._l1_ch.MatchUpstream(self._Ps_id, self._Bu_id)
        flg_2 = self._l2_ch.MatchUpstream(self._Ps_id, self._Bu_id)
        if flg_1 or flg_2:
            weight = 1.35

        return weight
    #---------------------------
    def _either_track_has(self, pid : int) -> bool:
        flg_l1 = self._l1_ch.HasInChain(pid)
        flg_l2 = self._l2_ch.HasInChain(pid)
        flg_kp = self._kp_ch.HasInChain(pid)

        return flg_l1 or flg_l2 or flg_kp
    #---------------------------
    def _get_psi_over_jpsi(self) -> float:
        flg_ps = self._either_track_has(self._Ps_id)
        if not flg_ps:
            return 1.0

        flg_bu = self._either_track_has(self._Bu_id)
        flg_bd = self._either_track_has(self._Bd_id)
        flg_bs = self._either_track_has(self._Bs_id)

        if flg_bs:
            return ( 5.40E-4/ 1.0800E-3 ) / ( 0.0748/ 0.1077) #0.72

        if flg_bu:
            return ( 6.19E-4/ 1.0006E-3 ) / ( 0.0729/ 0.1595) #1.35

        if flg_bd:
            return ( 5.90E-4/ 1.2700E-3 ) / (0.07610/ 0.1850) #1.13

        return 1.0
    #---------------------------
    def _get_kst_wgt(self) -> float:
        weight = 1.0
        if self._kp_ch.MatchDecay( [self._Pi_id, self._KS_id             ]):
            weight *= 0.5

        if self._kp_ch.MatchDecay( [self._Kp_id, self._Ph_id, self._Bd_id]):
            weight *= 0.5/0.9974

        if self._kp_ch.MatchDecay( [self._Kp_id, self._Ph_id, self._Bu_id]):
            weight *= 0.5/0.7597

        if self._kp_ch.MatchDecay( [self._Pi_id, self._Et_id] ):
            weight *= 0.28/0.4

        if self._kp_ch.MatchDecay( [self._Kp_id, self._Ks_id] ):
            weight *= 0.66 / 0.7993

        if self._kp_ch.MatchDecay( [self._Kp_id, self._Kst_c] ):
            weight *= 0.33 / 0.4993

        if self._kp_ch.MatchDecay( [self._Kp_id, self._K_2_1430_c]):
            weight *= 0.1670/0.2485

        return weight
    #---------------------------
    def get_weight(self) -> float:
        '''
        Returns:

        wt (float): Weight for candidate
        '''
        w1 = self._get_jpsi_wgt()
        w2 = self._get_brfrac_corr()
        w3 = self._get_psi_over_jpsi()
        w4 = self._get_kst_wgt()

        wt = w1 * w2 * w3 * w4

        return wt
    #---------------------------
    @staticmethod
    def get_chain(name, row : pnd.Series) -> PChain:
        '''
        Will return an instance of a PChain for a given dataframe row
        '''
        v1 = row[f'{name}_TRUEID']
        v2 = row[f'{name}_MC_MOTHER_ID']
        v3 = row[f'{name}_MC_GD_MOTHER_ID']
        v4 = row[f'{name}_MC_GD_GD_MOTHER_ID']

        chain = PChain(v1, v2, v3, v4)

        return chain
    #---------------------------
    @staticmethod
    def read_weight(row : pnd.Series, p1 : str, p2 : str, p3 : str) -> float:
        '''
        This method will return the BR weights

        Parameters
        -------------------
        row: Row in pandas dataframe holding PDG id information for the three particles.
        pX: Name of particle in dataframe, e.g. L1, L2, H

        Returns
        -------------------
        wgt (float): Weight for candidate associated to input row
        '''
        p1_ch = Reader.get_chain(p1, row)
        p2_ch = Reader.get_chain(p2, row)
        p3_ch = Reader.get_chain(p3, row)

        obj = Reader(p1_ch, p2_ch, p3_ch)
        wgt = obj.get_weight()

        return wgt
#---------------------------



'''
Module containing DataModel class
'''

from dmu.stats.zfit         import zfit
from dmu.generic            import utilities  as gut
from dmu.logging.log_store  import LogStore

from omegaconf              import DictConfig
from zfit                   import ComposedParameter
from zfit.core.interfaces   import ZfitPDF    as zpdf
from zfit.core.parameter    import Parameter  as zpar
from zfit.core.interfaces   import ZfitSpace  as zobs
from fitter.sim_fitter      import SimFitter

log = LogStore.add_logger('fitter:data_model')
# ------------------------
class DataModel:
    '''
    Model for fitting data samples
    '''
    # ------------------------
    def __init__(
            self,
            cfg     : DictConfig,
            obs     : zobs,
            trigger : str,
            project : str,
            q2bin   : str,
            name    : str|None=None):
        '''
        Parameters
        ------------------
        cfg    : Configuration object
        trigger: Hlt2RD...
        project: E.g. rx
        q2bin  : E.g. central
        obs    : zfit observable
        name   : Optional, identifier for this model
        '''
        self._cfg    = cfg
        self._obs    = obs
        self._trigger= trigger
        self._project= project
        self._q2bin  = q2bin
        self._name   = name
        self._nsig   = zfit.param.Parameter('nsignal', 100, 0, 1000_000)

        self._prc_pref = 'pscale' # This is the partially reconstructed scale preffix.
                                  # The name here mirrors what is in ConstraintReader.
    # ------------------------
    def _get_yield(self, name : str) -> zpar|ComposedParameter:
        '''
        Parameters
        --------------
        name: Sample name, e.g. Bu_Kee_eq_btosllball05_DPC.
              If component does not have MC associated, component name e.g. combinatorial

        Returns
        --------------
        zfit parameter used for extending it
        '''
        if name == 'signal':
            return self._nsig

        if name not in self._cfg.model.constraints.yields:
            log.debug(f'Yield for component {name} will be non-composed')
            return zfit.param.Parameter(f'n{name}', 100, 0, 1000_000)

        log.info(f'Yield for component {name} will be composed')
        # This scale should normally be below 1
        # It is nbackground / nsig
        # The parameter HAS TO start with pscale such that it is picked
        # by ConstraintReader
        scale= zfit.Parameter(f'{self._prc_pref}{name}', 0, 0, 10)
        nevt = zfit.ComposedParameter(f'n{name}', lambda x : x['nsig'] * x['scale'], params={'nsig' : self._nsig, 'scale' : scale})

        return nevt
    # ------------------------
    def _extend(self, pdf : zpdf, name : str) -> zpdf:
        '''
        Parameters
        -------------------
        name: Name of component
        pdf : zfit pdf

        Returns
        -------------------
        PDF with yield
        '''
        nevt = self._get_yield(name=name)

        kdes = zfit.pdf.KDE1DimFFT, zfit.pdf.KDE1DimExact, zfit.pdf.KDE1DimISJ
        if isinstance(pdf, kdes):
            pdf.set_yield(nevt)
            return pdf

        pdf = pdf.create_extended(nevt, name=name)

        return pdf
    # ------------------------
    def get_model(self) -> zpdf:
        '''
        Returns fitting model for data fit
        '''
        l_pdf = []
        npdf  = len(self._cfg.model)
        if npdf == 0:
            log.info(self._cfg.model)
            raise ValueError('Found zero components in model')

        log.debug(f'Found {npdf} components')
        for component, cfg_path in self._cfg.model.components.items():
            cfg = gut.load_conf(package='fitter_data', fpath=cfg_path)
            ftr = SimFitter(
                name     = self._name,
                component= component,
                trigger  = self._trigger,
                project  = self._project,
                q2bin    = self._q2bin,
                cfg      = cfg,
                obs      = self._obs)
            pdf = ftr.get_model()

            if pdf is None:
                log.warning(f'Skipping component: {component}')
                continue

            if component == 'signal':
                sample = component
            else:
                sample = cfg.get(key='sample', default_value=component)

            pdf    = self._extend(pdf=pdf, name=sample)
            l_pdf.append(pdf)

        pdf = zfit.pdf.SumPDF(l_pdf)

        return pdf
# ------------------------



'''
Script holding ConstraintReader class
'''
from dmu.logging.log_store       import LogStore
from rx_efficiencies.decay_names import DecayNames as dn
from fitter.prec_scales          import PrecScales

log=LogStore.add_logger('fitter:constraint_reader')
# -------------------------------------------------------------
class ConstraintReader:
    '''
    Class meant to provide constraints for fitting model
    '''
    # -------------------------------------------------------------
    def __init__(self, parameters : list[str], q2bin : str):
        '''
        Parameters: List of parameter names as in the PDF
        q2bin     : q2 bin
        '''

        self._l_par   = parameters
        self._q2bin   = q2bin

        self._d_const = {}
        self._signal  = 'bpkpee' # This is the signal decay nickname, needed for PRec scales constraints
        self._prc_pref= 'pscale' # This is the partially reconstructed scale preffix.
                                 # Yields are of the form: pscale{SAMPLENAME}
                                 # pscale needs to be removed to access sample name
    # -------------------------------------------------------------
    def _add_signal_constraints(self) -> None:
        raise NotImplementedError('This needs to be implemented with DataFitter')
    # -------------------------------------------------------------
    def _proc_from_par(self, par_name : str) -> str:
        sample = par_name.lstrip(self._prc_pref)
        decay  = dn.nic_from_sample(sample)

        return decay
    # -------------------------------------------------------------
    def _add_prec_constraints(self) -> None:
        for par in self._l_par:
            if not par.startswith('pscale'): # PRec constraints are scales, starting with "s"
                continue

            log.debug(f'Adding constrint for: {par}')

            process  = self._proc_from_par(par)
            obj      = PrecScales(proc=process, q2bin=self._q2bin)
            val, err = obj.get_scale(signal=self._signal)

            self._d_const[par] = val, err
    # -------------------------------------------------------------
    def get_constraints(self) -> dict[str,tuple[float,float]]:
        '''
        Returns dictionary with constraints, i.e.

        Key  : Name of fitting parameter
        Value: Tuple with mu and error
        '''
        self._add_prec_constraints()

        return self._d_const
# -------------------------------------------------------------



'''
Module containing DataFitter class
'''
from typing import cast

from omegaconf                import DictConfig, OmegaConf

from dmu.workflow.cache       import Cache
from dmu.stats.zfit           import zfit
from dmu.stats                import utilities  as sut
from dmu.logging.log_store    import LogStore
from rx_selection             import selection  as sel

from zfit.core.interfaces     import ZfitPDF    as zpdf
from zfit.core.interfaces     import ZfitSpace  as zobs
from fitter.data_preprocessor import DataPreprocessor
from fitter.base_fitter       import BaseFitter
from fitter.data_model        import DataModel
from fitter.constraint_reader import ConstraintReader

log=LogStore.add_logger('fitter:data_fitter')
# ------------------------
class DataFitter(BaseFitter, Cache):
    '''
    Fitter for data
    '''
    # ------------------------
    def __init__(
            self,
            sample  : str,
            trigger : str,
            project : str,
            q2bin   : str,
            cfg     : DictConfig,
            name    : str|None = None):
        '''
        name   : Identifier for fit, e.g. block. This is optional
        cfg    : configuration for the fit as a DictConfig object
        sample : Identifies sample e.g. DATA_24_MagUp...
        trigger: Hlt2RD...
        project: E.g. rx
        q2bin  : E.g. central
        cfg    : Configuration for the fit to data
        '''
        BaseFitter.__init__(self)

        self._sample    = sample
        self._trigger   = trigger
        self._project   = project
        self._q2bin     = q2bin
        self._cfg       = cfg
        self._name      = name
        self._base_path = self._get_base_path()

        Cache.__init__(
            self,
            out_path = self._base_path,
            cuts     = sel.selection(process=sample, trigger=trigger, q2bin=q2bin),
            config   = OmegaConf.to_container(cfg, resolve=True))

        self._obs = self._make_observable()
    # ------------------------
    def _get_base_path(self) -> str:
        '''
        Returns directory where outputs will go
        '''
        sample = self._sample.replace('*', 'p')
        if self._name is not None:
            sample = f'{self._cfg.output_directory}/{sample}/{self._name}/{self._trigger}_{self._project}_{self._q2bin}'
        else:
            sample = f'{self._cfg.output_directory}/{sample}/{self._trigger}_{self._project}_{self._q2bin}'

        return sample
    # ------------------------
    def _make_observable(self) -> zobs:
        '''
        Will return zfit observable
        '''
        name        = self._cfg.model.observable.name
        [minx, maxx]= self._cfg.model.observable.range

        return zfit.Space(name, limits=(minx, maxx))
    # ------------------------
    def _constraints_from_model(self, model : zpdf) -> dict[str,tuple[float,float]]:
        '''
        Parameters
        ----------------
        model: Model needed to fit the data

        Returns
        ----------------
        Dictionary with:

        Key: Name of parameter
        Value: Tuple value, error. Needed to apply constraints
        '''
        log.info('Getting constraints')

        s_par   = model.get_params()
        l_par   = [ par.name for par in s_par ]
        obj     = ConstraintReader(parameters = l_par, q2bin=self._q2bin)
        d_cns   = obj.get_constraints()

        log.debug(90 * '-')
        log.debug(f'{"Name":<20}{"Value":<20}{"Error":<20}')
        log.debug(90 * '-')
        for name, (val, err) in d_cns.items():
            log.debug(f'{name:<50}{val:<20.3f}{err:<20.3f}')
        log.debug(90 * '-')

        return d_cns
    # ------------------------
    def run(self) -> DictConfig:
        '''
        Runs fit

        Returns
        ------------
        DictConfig object with fitting results
        '''

        result_path = f'{self._out_path}/parameters.yaml'
        if self._copy_from_cache():
            res = OmegaConf.load(result_path)
            res = cast(DictConfig, res)

            return res

        dpr  = DataPreprocessor(
            obs    = self._obs,
            q2bin  = self._q2bin,
            sample = self._sample,
            trigger= self._trigger,
            out_dir= self._base_path,
            project= self._project)
        data = dpr.get_data()

        mod  = DataModel(
            name   = self._name,
            cfg    = self._cfg,
            obs    = self._obs,
            q2bin  = self._q2bin,
            trigger= self._trigger,
            project= self._project)
        model= mod.get_model()
        d_cns= self._constraints_from_model(model=model)

        res  = self._fit(data=data, model=model, d_cns=d_cns, cfg=self._cfg.fit)
        self._save_fit(
            cuts     = sel.selection(process=self._sample, trigger=self._trigger, q2bin=self._q2bin),
            cfg      = self._cfg.plots,
            data     = data,
            model    = model,
            res      = res,
            d_cns    = d_cns,
            out_path = self._out_path)

        cres = sut.zres_to_cres(res=res)

        self._cache()

        return cres
# ------------------------
